namespace recognition {

class Recognition {
public:
	class Builder : public IBuilder<Recognition> {
	private:
		int M;
		int N;
		int m;
		int n;
		unsigned char *camera_elements_uc;
		unsigned char *pattern_elements_uc;
	public:
		Builder& CameraHeight(int camera_height) {
			this->M = camera_height;
			return *this;
		}

		Builder& CameraWidth(int camera_width) {
			this->N = camera_width;
			return *this;
		}

		Builder& PatternHeight(int pattern_height) {
			this->m = pattern_height;
			return *this;
		}

		Builder& PatternWidth(int pattern_width) {
			this->n = pattern_width;
			return *this;
		}

		Builder& Camera_elements_uc(unsigned char *camera_elements_uc) {
			this->camera_elements_uc = camera_elements_uc;
			return *this;
		}

		Builder& Pattern_elements_uc(unsigned char *pattern_elements_uc) {
			this->pattern_elements_uc = pattern_elements_uc;
			return *this;
		}

		Recognition build() {
			return Recognition(this);
		}

		friend class Recognition;
	};

	static const log4cxx::LoggerPtr logger;

	static cv::Mat rotateImage(const cv::Mat& source, double angle);

	void recognize(int &x_out, int &y_out);

private:
	Recognition(Builder* builder) {
		this->M = builder->M;
		this->N = builder->N;
		this->m = builder->m;
		this->n = builder->n;
		this->camera_elements_uc = builder->camera_elements_uc;
		this->pattern_elements_uc = builder->pattern_elements_uc;
	}

	int M;			// Camera height
	int N;			// Camera width
	int m;			// Pattern height
	int n;			// Pattern width
	unsigned char *camera_elements_uc;
	unsigned char *pattern_elements_uc;
};

}	// namespace
