#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <iostream>

#include <log4cxx/logger.h>
#include <log4cxx/xml/domconfigurator.h>
#include <base/logging.h>

#include <cuda_runtime.h>
#include <helper_cuda.h>
#include <helper_functions.h>

#include <cv.h>
#include <highgui.h>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/core/core.hpp"
#include "opencv2/highgui/highgui.hpp"

#include <base/types.cuh>
#include <recognition.cuh>

#include <gpu/fast_match_helper.cuh>
#include <gpu/fast_match.cuh>

#include <gpu/ciratefi_helper.cuh>
#include <gpu/ciratefi.cuh>

#include <cpu/helper.h>
#include <debug/garbage.cuh>
#include <debug/simple_test.cuh>
#include <main.cuh>

using namespace std;
using namespace cv;
using namespace recognition;

// Main Function
int main(int argc, const char * argv[]) {
//	for (double d = 0.25; d > 0.01; d -= 0.01) {
//		printf("d=%f, res=%f\n", d, sqrt(pow(d, 3)) * 128);
//	}
//	exit(0);
//
//	double delta_ = 0.25;
//	int delta_int = 25;
//	int divisor = 100;
//
//	printf("1/d3=%f 1/d4=%f\n", (double) pow(divisor, 3)/pow(delta_int, 3), (double) pow(divisor, 4)/pow(delta_int, 4));
//	exit(0);

	// Testing video recognition real time
//	video_test();
//	exit(0);

	// Configuring logging
	string configFileStr(CONFIG_DIRECTORY);
	configFileStr.push_back(FILE_PATH_SEPARATOR);
	configFileStr.append(CONFIG_FILE_XML);
	log4cxx::xml::DOMConfigurator::configure(configFileStr.c_str());

	// Get default logger
	log4cxx::LoggerPtr logger(log4cxx::Logger::getRootLogger());
	LOG_INFO(logger, "Entering application.");
	LOG_INFO(logger, "Reading command line arguments...");

	if (!argv[1] || !argv[2]) {
		LOG_ERROR(logger, "Both input files must be specified!");
		exit(EXIT_FAILURE);
	}

	string imagesDirStr(RESOURCES_DIRECTORY);
	imagesDirStr.push_back(FILE_PATH_SEPARATOR);

	// Input camera and pattern images' file names
	string cameraFileStr(imagesDirStr + argv[1]);
	string patternFileStr(imagesDirStr + argv[2]);

	unsigned char *camera_elements_uc;		// camera image bytes
	unsigned char *pattern_elements_uc;		// pattern image bytes

	// Read the data and convert it into gray scale
//	camera_elements_uc = BmpRgbToGrayscale(ReadBmpRgb(cameraFileStr.c_str(), M, N), M, N);
//	pattern_elements_uc = BmpRgbToGrayscale(ReadBmpRgb(patternFileStr.c_str(), m, n), m, n);

	cv::Mat cam = cv::imread(cameraFileStr.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
	cv::Mat pat = cv::imread(patternFileStr.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
	camera_elements_uc = cam.ptr(0);
	pattern_elements_uc = pat.ptr(0);
	//%% image dimensions
	int h1, w1, h2, w2;
	h2 = cam.rows; w2 = cam.cols;
	h1 = pat.rows; w1 = pat.cols;

	LOG4CXX_INFO(logger, kInputArgumentsTableBorders);
	LOG_INFO(logger, "|%-15s|%-40s|%-10s|%-10s|", "Image", "File", "Width", "Height");
	LOG4CXX_INFO(logger, kInputArgumentsTableBorders);
	LOG_INFO(logger, "|%-15s|%-40s|%-10d|%-10d|", "camera", cameraFileStr.c_str(), w2, h2);
	LOG_INFO(logger, "|%-15s|%-40s|%-10d|%-10d|", "pattern", patternFileStr.c_str(), w1, h1);
	LOG4CXX_INFO(logger, kInputArgumentsTableBorders);
	LOG_INFO(logger, "Target (x, y)  =  (%d, %d)", -1, -1);	// TODO: Opt is unspecified

	// Not cropped images
	double *_image_elements, *_template_elements;
	_image_elements = UcharToNormalizedDouble(camera_elements_uc, h2, w2);
	_template_elements = UcharToNormalizedDouble(pattern_elements_uc, h1, w1);

	// Cropped images with odd dimensions
	double *image_elements, *template_elements;

	image_elements = MakeOdd(_image_elements, h2, w2);
	template_elements = MakeOdd(_template_elements, h1, w1);

	free(_image_elements);
	free(_template_elements);

	cv::Mat image(h2, w2, CV_64FC1, image_elements);
	cv::Mat pattern(h1, w1, CV_64FC1, template_elements);

//	DEBUG_Recolor(pattern);
//	DEBUG_Recolor(image);

	int cuda_device = 0;
	checkCudaErrors(cudaSetDevice(cuda_device));
	cudaDeviceProp deviceProp;
	checkCudaErrors(cudaGetDeviceProperties(&deviceProp, cuda_device));
	int max_shared_memory = deviceProp.sharedMemPerBlock;

///////////////////////////////// Ciratefi /////////////////////////////////

	double blur_sigma = 4;
	double blur_size = 2 * blur_sigma - 1;

	//TODO:
	Mat I1_ciratefi, I2_ciratefi;
	I1_ciratefi = pattern.clone();
	I2_ciratefi = image.clone();

	// Smooth
	GaussianBlur(I1_ciratefi, I1_ciratefi, Size(blur_size, blur_size), blur_sigma, blur_sigma, BORDER_REFLECT);
	GaussianBlur(I2_ciratefi, I2_ciratefi, Size(blur_size, blur_size), blur_sigma, blur_sigma, BORDER_REFLECT);

	BCThresholds bc_thresholds(0.1, 1);
	CiratefiThresholds c_thresholds(0.95, 0.9, 0.9);

	CiratefiBoundaries c_bounds;
	c_bounds.s = BoundsWithStep<double>(0.1, 0.4, 1.3);
	c_bounds.r = BoundsWithStep<int>(2, 0, min(I1_ciratefi.cols / 2, I1_ciratefi.rows / 2) - c_bounds.r.step);
	c_bounds.phi = BoundsWithStep<double>(2 * M_PI / 36, 0, 2 * M_PI);
	c_bounds.angle_incr = 2;
	c_bounds.line_incr = 2;

	// TODO:
	recognition::gpu::Ciratefi ciratefi = recognition::gpu::Ciratefi::Builder()
		.I1(I1_ciratefi)
		.I2(I2_ciratefi)
		.BC_thresholds(bc_thresholds)
		.thresholds(c_thresholds)
		.bounds(c_bounds)
		.maximum_shared_memory(max_shared_memory)
		.build();

//	DEBUG_RunCiratefi(ciratefi);
//	exit(0);

//	free(image_elements);
//	free(template_elements);

///////////////////////////////// FAsT-Match /////////////////////////////////

	// Set default values for optional variables
	double epsilon, delta, photometricInvariance, minScale, maxScale;
	epsilon = 0.15;
	delta = 0.25;
	photometricInvariance = 1;
	minScale = 0.5;
	maxScale = 2;

	double r1x, r1y, r2x, r2y;
	r1x = 0.5 * (w1-1);
	r1y = 0.5 * (h1-1);
	r2x = 0.5 * (w2-1);
	r2y = 0.5 * (h2-1);

	//%% search range
	double minTx, maxTx, minTy, maxTy, minR, maxR;
	minTx = -(r2x - r1x * minScale);
	maxTx = r2x - r1x * minScale;
	minTy = -(r2y - r1y * minScale);
	maxTy = r2y - r1y * minScale;
	minR = -M_PI;
	maxR = M_PI;

	//%% parametrize the initial grid
	NetParameters bounds_min, bounds_max, steps;
	bounds_min.tx = minTx; bounds_max.tx = maxTx;
	bounds_min.ty = minTy; bounds_max.ty = maxTy;
	bounds_min.r = minR; bounds_max.r = maxR;
	bounds_min.s = minScale; bounds_max.s = maxScale;

	steps.tx = delta * w1 / sqrt(2);
	steps.ty = delta * h1 / sqrt(2);
	steps.r = delta * sqrt(2);
	steps.s = delta / sqrt(2);

	// TODO:
	recognition::gpu::BestTransformation bestTransformation = recognition::gpu::BestTransformation::Builder()
		.Template(pattern)
		.Image(image)
		.Bounds(bounds_min, bounds_max)
		.Steps(steps)
		.Epsilon(epsilon)
		.Delta(delta)
		.PhotometricInvariance(photometricInvariance)
		.build();

	BestTransformationResult bestTransformationResult = bestTransformation.FindBestTransformation();
	Mat config = bestTransformationResult.bestConfig;
	Mat affine = bestTransformationResult.bestTransMat.rowRange(0, 2);

	cout << "Best transformation:" << endl;
	cout << bestTransformationResult.bestConfig << endl;
	cout << bestTransformationResult.bestTransMat << endl;
	cout << bestTransformationResult.sampledError << endl;
	cout << "*******************************  TESTS **************************" << endl;

	double sx, sy;
	sx = config.at<double>(0, 3);
	sy = config.at<double>(0, 4);

	double a11, a12, a13, a21, a22, a23;
	a11 = affine.at<double>(0, 0);
	a12 = affine.at<double>(0, 1);
	a13 = affine.at<double>(0, 2);
	a21 = affine.at<double>(1, 0);
	a22 = affine.at<double>(1, 1);
	a23 = affine.at<double>(1, 2);

	Mat I1 = pattern;
	Mat I2 = image;
	Mat I1_ = Mat::zeros(I1.rows * sy, I1.cols * sx, I1.type());

	// Detect sizes of transformed pattern
	Mat corners = (Mat_<int>(2, 4) <<	0, a11 * (I1.cols - 1), a11 * (I1.cols - 1) + a12 * (I1.rows - 1), a12 * (I1.rows - 1),
										0, a21 * (I1.cols - 1), a21 * (I1.cols - 1) + a22 * (I1.rows - 1), a22 * (I1.rows - 1));
	cout << "corners:\n" << corners;

	int xDim, yDim;	// sizes of transformed pattern
	int minX, minY, maxX, maxY;	// min, max of components x and y
	minX = minY = INT_MAX;
	maxX = maxY = INT_MIN;

	for (int i = 0; i < 4; ++i) {
		int x, y;
		x = corners.at<int>(0, i);
		y = corners.at<int>(1, i);

		if (x < minX) minX = x;
		if (x > maxX) maxX = x;
		if (y < minY) minY = y;
		if (y > maxY) maxY = y;
	}

	xDim = maxX - minX + 1;
	yDim = maxY - minY + 1;

	cout << "DIMS: " << xDim << " " << yDim << endl;
	I1_ = Mat::zeros(yDim, xDim, I1.type());

	Point2f srcTri[3];
	Point2f dstTri[3];


//	/// Set your 3 points to calculate the  Affine Transform
//	srcTri[0] = Point2f(0, 0);
//	srcTri[1] = Point2f(I1.cols - 1, 0);
//	srcTri[2] = Point2f(0, I1.rows - 1);
//
//	dstTri[0] = Point2f(0, 0);
//	dstTri[1] = Point2f((int) (a11 * (I1.cols - 1)),
//			(int) (a21 * (I1.cols - 1)));
//	dstTri[2] = Point2f((int) (a12 * (I1.rows - 1)),
//			(int) (a22 * (I1.rows - 1)));
//
//	/// Get the Affine Transform
//	Mat warp_mat(2, 3, CV_32FC1);
//	warp_mat = getAffineTransform(srcTri, dstTri);

	Mat warp_mat = affine;
	warp_mat.at<double>(0, 2) = -minX;
	warp_mat.at<double>(1, 2) = -minY;
	cout << "warp_mat:\n" << warp_mat << endl;

	/// Apply the Affine Transform just found to the src image
	warpAffine(I1, I1_, warp_mat, I1_.size());

	namedWindow("Pattern", CV_WINDOW_AUTOSIZE);
	imshow("Pattern", I1);

	namedWindow("Transformed pattern", CV_WINDOW_AUTOSIZE);
	imshow("Transformed pattern", I1_);

	Mat I2_ = I2.clone();

	for (int y = 0; y < I1.rows; ++y) {
		for (int x = 0; x < I1.cols; ++x) {
			int x_, y_;
			x_ = a11 * (x - (r1x + 1)) + a12 * (y - (r1y + 1)) + a13 + (r2x + 1);
			y_ = a21 * (x - (r1x + 1)) + a22 * (y - (r1y + 1)) + a23 + (r2y + 1);
			I2_.at<double>(y_, x_) = I1.at<double>(y, x) * 0.8;
		}
	}

	namedWindow("Image", CV_WINDOW_AUTOSIZE);
	imshow("Image", I2_);
	while (waitKey(0) != 27);

//  // Display transformed pattern (output image is not smooth)
//	for (int y = 0; y < I1.rows; ++y) {
//		for (int x = 0; x < I1.cols; ++x) {
//			int x_, y_;
//			x_ = a11 * x + a12 * y - minX;
//			y_ = a21 * x + a22 * y - minY;
//			I1_.at<double>(y_, x_) = I1.at<double>(y, x);
//
//			if ((x == 0 && y == 0) || (x == I1.cols - 1 && y == 0) || (x == 0 && y == I1.rows - 1) || (x == I1.cols - 1 && y == I1.rows - 1))
//				printf("x,y = %d,%d: %d %d\n", x, y, int(a11 * x + a12 * y), int(a21 * x + a22 * y));
//
//		}
//	}
//	namedWindow( "trans", CV_WINDOW_AUTOSIZE );
//	imshow( "trans", I1_ );
//	waitKey(0);

	LOG4CXX_INFO(logger, "Exiting application...");
    return 0;
}
