#include <stdlib.h>
#include <stdio.h>
#include <cuda_runtime.h>

#include <log4cxx/logger.h>
#include <base/logging.h>

#include <cv.h>
#include <highgui.h>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/photo/photo.hpp"
#include "opencv2/core/core.hpp"

#include <base/types.cuh>
#include <base/methods.cuh>
#include <gpu/gpu_correlation2.cuh>

#include <gpu/fast_match_helper.cuh>
#include <gpu/fast_match.cuh>

#include <recognition.cuh>

namespace recognition {

const log4cxx::LoggerPtr Recognition::logger = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("recognition.Recognition"));

cv::Mat Recognition::rotateImage(const cv::Mat& source, double angle) {
	cv::Point2f src_center(source.cols/2.0F, source.rows/2.0F);
	cv::Mat rot_mat = cv::getRotationMatrix2D(src_center, angle, 1.0);
	cv::Mat dst;
	cv::warpAffine(source, dst, rot_mat, source.size());
	return dst;
}

void Recognition::recognize(int &x_out, int &y_out) {
//		cv::Mat img(M, N, CV_8UC1, camera_uc, N);
//		cv::Mat pat(m, n, CV_8UC1, pattern_uc, n);
//
//		cv::GaussianBlur(img, img, cv::Size(7,7), 1.5, 1.5);
//
//		cv::namedWindow( "Camera blur", CV_WINDOW_AUTOSIZE );
//		cv::imshow( "Camera blur", img );
//
//		cv::fastNlMeansDenoising( img, img, 9.0f, 7, 21 );
//
//		cv::namedWindow( "Camera blur denoising", CV_WINDOW_AUTOSIZE );
//		cv::imshow( "Camera blur denoising", img );
//
//		cv::resize(pat, pat, cv::Size(512, 512));
//
//		cv::GaussianBlur(pat, pat, cv::Size(3,3), 0, 0);
//		cv::fastNlMeansDenoising( pat, pat, 9.0f, 7, 21 );
//
//		int x_res = 356;
//		int y_res = 137;
//		cv::rectangle(img, cv::Point(x_res, y_res), cv::Point(x_res + n, y_res + m), cv::Scalar_<double>(0), 2);
//
//		cv::namedWindow( "Camera", CV_WINDOW_AUTOSIZE );
//		cv::imshow( "Camera", img );
//
//		cv::Mat rot = rotateImage(pat, 60);
//		cv::namedWindow( "Pattern", CV_WINDOW_AUTOSIZE );
//		cv::imshow( "Pattern", rot );
//
//		cv::waitKey(0);
//		exit(0);

	// Matrix representation of camera
	Matrix2uchar m_camera_uc;
	m_camera_uc.height = M;
	m_camera_uc.width = m_camera_uc.stride = N;
	m_camera_uc.elements = camera_elements_uc;

	// Matrix representations of pattern
	Matrix2uchar m_pattern_uc;
	m_pattern_uc.height = m;
	m_pattern_uc.width = m_pattern_uc.stride = n;
	m_pattern_uc.elements = pattern_elements_uc;

	// Matrix of camera with img_t elements
	Matrix2f camera;
	camera.height = m_camera_uc.height;
	camera.width = camera.stride = m_camera_uc.stride;
	int cam_size = m_camera_uc.height * m_camera_uc.width;
	camera.elements = (img_t *) malloc(sizeof(img_t) * cam_size);
	for (int i = 0; i < cam_size; ++i) {
		camera.elements[i] = (img_t) camera_elements_uc[i];
	}

	// Integral representation of camera elements and squared camera elements
	int *cam_integral = GetIntegralRepr<int, unsigned char>(camera_elements_uc, M, N);
	long *cam_integral2 = GetIntegralReprSquared<long, unsigned char>(camera_elements_uc, M, N);

	//////////// VIDEO /////////////
	gpu::GPUCorrelation2 gpuCorrelation2 = gpu::GPUCorrelation2::Builder()
		.Camera(camera)
		.M_pattern_uc(m_pattern_uc)
		.build();

	gpuCorrelation2.find(1, 0.9F, x_out, y_out);
	////////////////////////////////

//	cv::Mat cam(M, N, CV_8UC1);
//	cv::Mat pat(m, n, CV_8UC1);
//	cam.data = camera_elements_uc;
//	pat.data = pattern_elements_uc;

//	CV_64FC1

//	imshow("cam", cam);
//	imshow("pat", pat);
//
//	for (;;) {
//		if (cv::waitKey(30) >= 0) {
//			break;
//		}
//	}

	LOG_INFO(logger, "Freeing resources...");
	free(cam_integral);
	free(cam_integral2);
	free(camera.elements);
}

}	// namespace recognition
