namespace recognition {
namespace cpu {

class CPUCorrelation {
public:
	class Builder : public IBuilder<CPUCorrelation> {
	public:
		Builder& Camera(Matrix2f &camera) {
			this->camera = camera;
			return *this;
		}

		Builder& M_pattern_uc(Matrix2uchar &m_pattern_uc) {
			this->m_pattern_uc = m_pattern_uc;
			return *this;
		}

		Builder& Camera_integral(int* &camera_integral) {
			this->camera_integral = camera_integral;
			return *this;
		}

		Builder& Camera_integral2(long* &camera_integral2) {
			this->camera_integral2 = camera_integral2;
			return *this;
		}

		Builder& X_target(int x_target) {
			this->x_target = x_target;
			return *this;
		}

		Builder& Y_target(int y_target) {
			this->y_target = y_target;
			return *this;
		}

		CPUCorrelation build() {
			return CPUCorrelation(this);
		}

	private:
		Matrix2f camera;
		Matrix2uchar m_pattern_uc;
		int *camera_integral;
		long *camera_integral2;
		int x_target;
		int y_target;

		friend class CPUCorrelation;
	};

	static const log4cxx::LoggerPtr logger;
	static const log4cxx::LoggerPtr output;

	CorrelationResult FindCorrelation(Matrix2f pattern,
			int *stoch_vect, unsigned v_length, img_t pattern_sum, unsigned n_elements);

	// Estimating average time and deviation with fixed probability
	void average(unsigned samples = 1, float from = 0.0F, float to = 1.0F);

private:
	CPUCorrelation(Builder* builder) {
		this->camera = builder->camera;
		this->m_pattern_uc = builder->m_pattern_uc;
		this->camera_integral = builder->camera_integral;
		this->camera_integral2 = builder->camera_integral2;
		this->x_target = builder->x_target;
		this->y_target = builder->y_target;

		iteration_ = 0;
		skip_ = 0.0F;
	}

	Matrix2f camera;
	Matrix2uchar m_pattern_uc;
	int *camera_integral;
	long *camera_integral2;
	int x_target;
	int y_target;

	int iteration_;
	float skip_;
};

}	// namespace cpu
}	// namespace recognition
