namespace recognition {
namespace cpu {

class CPUCorrelation2 {
public:
	class Builder : public IBuilder<CPUCorrelation2> {
	public:
		Builder& Camera(Matrix2f &camera) {
			this->camera = camera;
			return *this;
		}

		Builder& M_pattern_uc(Matrix2uchar &m_pattern_uc) {
			this->m_pattern_uc = m_pattern_uc;
			return *this;
		}

		Builder& X_target(int x_target) {
			this->x_target = x_target;
			return *this;
		}

		Builder& Y_target(int y_target) {
			this->y_target = y_target;
			return *this;
		}

		CPUCorrelation2 build() {
			return CPUCorrelation2(this);
		}

	private:
		Matrix2f camera;
		Matrix2uchar m_pattern_uc;
		int x_target;
		int y_target;

		friend class CPUCorrelation2;
	};

	static const log4cxx::LoggerPtr logger;
	static const log4cxx::LoggerPtr output;

	CorrelationResult FindCorrelation(SparseImage pattern, img_t pattern_sum);

	// Estimating average time and deviation with fixed probability
	void average(unsigned samples = 1, float from = 0.0F, float to = 1.0F);

private:
	CPUCorrelation2(Builder* builder) {
		this->camera = builder->camera;
		this->m_pattern_uc = builder->m_pattern_uc;
		this->x_target = builder->x_target;
		this->y_target = builder->y_target;

		iteration_ = 0;
		skip_ = 0.0F;
	}

	Matrix2f camera;
	Matrix2uchar m_pattern_uc;
	int x_target;
	int y_target;

	int iteration_;
	float skip_;
};

}	// namespace cpu
}	// namespace recognition
