// For the CUDA runtime routines (prefixed with "cuda_")
#include <cuda_runtime.h>

namespace recognition {
namespace gpu {

struct _cu_ImgCorrArgs {
	_cu_ImgCorrArgs(int *_d_integral, long *_d_integral2, int *_d_stoch_vect, unsigned _v_length, img_t _pattern_sum, unsigned _n_elements):
		d_integral(_d_integral), d_integral2(_d_integral2), d_stoch_vect(_d_stoch_vect),
		v_length(_v_length), pattern_sum(_pattern_sum), n_elements(_n_elements), debug(0) {}

	int *d_integral;
	long *d_integral2;
	int *d_stoch_vect;
	unsigned v_length;
	img_t pattern_sum;
	unsigned n_elements;
	float debug;
};

// Thread block size
#define SUBMATRIX_SIZE (2 * BLOCK_SIZE - 1)

class GPUCorrelation {
public:
	class Builder : public IBuilder<GPUCorrelation> {
	public:
		Builder& Camera(Matrix2f &camera) {
			this->camera = camera;
			return *this;
		}

		Builder& M_pattern_uc(Matrix2uchar &m_pattern_uc) {
			this->m_pattern_uc = m_pattern_uc;
			return *this;
		}

		Builder& Camera_integral(int* &camera_integral) {
			this->camera_integral = camera_integral;
			return *this;
		}

		Builder& Camera_integral2(long* &camera_integral2) {
			this->camera_integral2 = camera_integral2;
			return *this;
		}

		Builder& X_target(int x_target) {
			this->x_target = x_target;
			return *this;
		}

		Builder& Y_target(int y_target) {
			this->y_target = y_target;
			return *this;
		}

		GPUCorrelation build() {
			return GPUCorrelation(this);
		}

	private:
		Matrix2f camera;
		Matrix2uchar m_pattern_uc;
		int *camera_integral;
		long *camera_integral2;
		int x_target;
		int y_target;

		friend class GPUCorrelation;
	};

	static const log4cxx::LoggerPtr logger;
	static const log4cxx::LoggerPtr output;

	CorrelationResult FindCorrelation(Matrix2f *d_camera, Matrix2f pattern, _cu_ImgCorrArgs *d_args);

	// Estimating average time and deviation with fixed probability
	void average(unsigned samples = 1, float from = 0.0F, float to = 1.0F);

private:
	GPUCorrelation(Builder* builder) {
		this->camera = builder->camera;
		this->m_pattern_uc = builder->m_pattern_uc;
		this->camera_integral = builder->camera_integral;
		this->camera_integral2 = builder->camera_integral2;
		this->x_target = builder->x_target;
		this->y_target = builder->y_target;

		iteration_ = 0;
		skip_ = 0.0F;
	}

	Matrix2f camera;
	Matrix2uchar m_pattern_uc;
	int *camera_integral;
	long *camera_integral2;
	int x_target;
	int y_target;

	int iteration_;
	float skip_;
};

__global__ void _cu_fullCorrelationKernel(Matrix2f *A, Matrix2f *B, CorrelationPoint *C, _cu_ImgCorrArgs *args);

}	// namespace gpu
}	// namespace recognition
