#include <cuda_runtime.h>

namespace recognition {
namespace gpu {

class GPUCorrelation2 {
public:
	class Builder : public IBuilder<GPUCorrelation2> {
	public:
		Builder& Camera(Matrix2f &camera) {
			this->camera = camera;
			return *this;
		}

		Builder& M_pattern_uc(Matrix2uchar &m_pattern_uc) {
			this->m_pattern_uc = m_pattern_uc;
			return *this;
		}

		Builder& X_target(int x_target) {
			this->x_target = x_target;
			return *this;
		}

		Builder& Y_target(int y_target) {
			this->y_target = y_target;
			return *this;
		}

		GPUCorrelation2 build() {
			return GPUCorrelation2(this);
		}

	private:
		Matrix2f camera;
		Matrix2uchar m_pattern_uc;
		int x_target;
		int y_target;

		friend class GPUCorrelation2;
	};

	static const log4cxx::LoggerPtr logger;
	static const log4cxx::LoggerPtr output;

	CorrelationResult FindCorrelation(Matrix2f *d_camera, SparseImage pattern, img_t pattern_sum, int &x_out, int &y_out);

	// Estimating average time and deviation with fixed probability
	void average(unsigned samples = 1, float from = 0.0F, float to = 1.0F);

	void find(unsigned samples, float skip_probability, int &x_out, int &y_out);

private:
	GPUCorrelation2(Builder* builder) {
		this->camera = builder->camera;
		this->m_pattern_uc = builder->m_pattern_uc;
		this->x_target = builder->x_target;
		this->y_target = builder->y_target;

		iteration_ = 0;
		skip_ = 0.0F;
	}

	Matrix2f camera;
	Matrix2uchar m_pattern_uc;
	int x_target;
	int y_target;

	int iteration_;
	float skip_;
};

struct _cu_findCorrelationKernelArgs {
	_cu_findCorrelationKernelArgs(img_t _pattern_sum):
		pattern_sum(_pattern_sum), debug(0.0F) {}

	img_t pattern_sum;
	float debug;
};

__global__ void _cu_findCorrelationKernel(Matrix2f *A, SparseImage *B, CorrelationPoint *C, _cu_findCorrelationKernelArgs *args);

// DEPRECATED
__global__ void _cu_findMaximumKernel(CorrelationPoint *g_idata, CorrelationPoint *g_odata);

}	// namespace gpu
}	// namespace recognition
