#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <log4cxx/logger.h>
#include <log4cxx/xml/domconfigurator.h>
#include <base/logging.h>

#include <cv.h>
#include <highgui.h>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/photo/photo.hpp"
#include "opencv2/core/core.hpp"

#include <base/types.cuh>
#include <recognition.cuh>

#include <video.cuh>

using namespace cv;
using namespace recognition;

void video_test() {
	VideoCapture cap("resources/long.mp4");
	if (!cap.isOpened())  // check if we succeeded
		exit(1);

	double cap_width = cap.get(CV_CAP_PROP_FRAME_WIDTH);
	printf("width: %lf\n", cap_width);

	double cap_height = cap.get(CV_CAP_PROP_FRAME_HEIGHT);
	printf("height: %lf\n", cap_height);


//	exit(0);

	int n_frames = 0;
	const int M = 384;
	const int N = 512;
	const int m = 128;
	const int n = 128;
	const int y_target = 0;
	const int x_target = 0;

	Mat pat = imread("images/ball-long.bmp", CV_LOAD_IMAGE_GRAYSCALE);
	unsigned char *pattern_elements_uc = pat.ptr(0);

	Mat img;
	namedWindow("Image", 1);
	for (;;) {
		Mat frame;
		cap >> frame; // get a new frame from camera
		cvtColor(frame, img, CV_RGB2GRAY);
//		edges = frame;
		Rect crop(0, 0, 960, 720);
		img = img(crop);
		resize(img, img, Size(512, 384));

		unsigned char *camera_elements_uc = img.ptr(0);

		int x_out = 0;
		int y_out = 0;
		if (n_frames++ % 1 == 0) {
			Recognition recognition = Recognition::Builder()
				.CameraHeight(M)
				.CameraWidth(N)
				.PatternHeight(m)
				.PatternWidth(n)
				.X_target(x_target)
				.Y_target(y_target)
				.Camera_elements_uc(camera_elements_uc)
				.Pattern_elements_uc(pattern_elements_uc)
				.build();

			recognition.recognize(x_out, y_out);

		float c = 1.875;

//		rectangle(img, Point(x_out, y_out), Point(x_out + n, y_out + m), Scalar_<double>(0), 2);
		rectangle(frame, Point(c * x_out, c * y_out), Point(c * (x_out + n), c * (y_out + m)), Scalar_<double>(0), 2);
		rectangle(frame, Point(960, 0), Point(1280, 720), Scalar_<double>(0), CV_FILLED);

//		line(frame, Point(960, 0), Point(960, 720), Scalar_<double>(0), 300);

//		GaussianBlur(img, img, Size(7, 7), 1.5, 1.5);
//		Canny(img, img, 0, 30, 3);
//		imshow("Image", img);
		imshow("Image", frame);
		if (waitKey(30) >= 0)
			break;
		}
	}

	printf("\nframes processed: %d", n_frames);
}
